unit TimedMessageDialog;

interface

uses dialogs;

function timedMessageDlg(const Msg: string; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: Longint; AInterval: integer=10):integer;

implementation

uses ExtCtrls, Contnrs, forms, Classes, StdCtrls, sysutils, windows
{$IFDEF madExcept}
     , madStackTrace
{$ENDIF}
;

type
  TDialogHook=class
  private
    timer, catcher:TTimer;
    forms:TObjectList;
    oldCount:integer;
    form:TForm;
    defaultButton:TButton;
    originalButtontext:string;
    remaining:integer;
    trace:string;
    procedure OnTimer(Sender:TObject);
    procedure OnCatcherTimer(Sender:TObject);
    procedure displayTime;
    procedure traceHack(Sender: TObject; var Key: Word; Shift: TShiftState);
  public
    constructor Create(AInterval:integer=10);
    destructor Destroy; override; 
  end;

{ TDialogHook }

constructor TDialogHook.Create(AInterval: integer);
var i:integer;
begin
  {$IFDEF madExcept}
  trace:=StackTrace(false, true, true);
  {$ENDIF}
  remaining:=AInterval-1;
  timer:=TTimer.Create(nil);
  timer.Enabled:=false;
  timer.Interval:=1000;
  timer.OnTimer:=OnTimer;

  catcher:=TTimer.Create(nil);
  catcher.Enabled:=false;
  catcher.Interval:=1000;
  catcher.OnTimer:=OnCatcherTimer;

  forms:=TObjectList.Create(false);
  for i:=1 to application.ComponentCount do
    if application.Components[i-1] is TForm then
      forms.add(application.Components[i-1]);
  oldCount:=application.ComponentCount;
  catcher.Enabled:=true;
end;

destructor TDialogHook.Destroy;
begin
  forms.free;
  catcher.free;
  timer.free;
  inherited;
end;

procedure TDialogHook.displayTime;
begin
  defaultButton.caption:=originalButtontext+' ['+inttostr(remaining)+' s]'
end;

procedure TDialogHook.OnCatcherTimer(Sender: TObject);
  function isTheNewDialog(c:TComponent):boolean;
  begin
    result:=(c is TForm) and (c.ClassName='TMessageForm') and (forms.IndexOf(c)=-1){this shouldn't be necessary but just to be safe};
  end;
var i:integer;
begin
  catcher.Enabled:=false;
  assert(oldCount=application.ComponentCount-1);
  i:=0;
  while (i<application.ComponentCount) and (not isTheNewDialog(application.Components[i])) do
    inc(i);
  assert(i<application.ComponentCount);
  form:=application.Components[i] as TForm;
  form.OnKeyUp:=traceHack;
  form.KeyPreview:=true;
  i:=0;
  while (i<form.ComponentCount) and ((not (form.Components[i] is TButton)) or (TButton(form.Components[i]).Default=false)) do
    inc(i);
  assert(i<form.ComponentCount);
  defaultButton:=TButton(form.Components[i]);
  originalButtontext:=defaultButton.caption;
  displayTime;
  timer.enabled:=true;
end;

procedure TDialogHook.OnTimer(Sender: TObject);
begin
  dec(remaining);
  if remaining>-1 then displayTime
                  else defaultButton.Click;
end;

function timedMessageDlg(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons; HelpCtx: Longint; AInterval: integer): Integer;
var dh:TDialogHook;
begin
  dh:=TDialogHook.Create(AInterval);
  Result := MessageDlg(Msg, DlgType, Buttons, HelpCtx);
  dh.Free;
end;

procedure TDialogHook.traceHack(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (ssctrl in shift) and (ssshift in shift) and (key=ord('T')) then
  begin
    timer.Enabled:=false;
    showmessage(trace);
    timer.Enabled:=true;
  end;
end;

end.
